package net.chlup.mybrother;

import java.util.*;

import android.app.Service;
import android.app.AlarmManager;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Notification;

import android.os.Bundle;
import android.os.IBinder;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.BroadcastReceiver;

import android.location.*;
import android.telephony.TelephonyManager;
import android.util.Base64;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.text.SimpleDateFormat;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.NameValuePair;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.client.DefaultHttpRequestRetryHandler;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;

public class GPS extends Service {
    /* Location manager params */
    LocationManager lm;
    LocListen ll;
    AlarmManager am;
    PendingIntent amintent;
    LocAlarm recvr;
    PendingIntent tointent;
    LocTimeout recvTimeout;
    boolean twiceTimeout;
    long timeout;

    /* Prepare connection and request */
    DefaultHttpClient httpclient; // presunuto sem nahoru z "postLocation()", aby fungoval keep-alive pri rychlejsim zasilani souradnic
    int waiting_for_first_send_data = 1; 

    /* Pole se zjistenymi udaji */
    long numberPoint = 0;
    long numberPointSends = 0;
	List<NameValuePair> nameValuePairs;

    /* Date formatting */
    SimpleDateFormat dateformatter;

    /* Notification */
    NotificationManager notman;
    Notification notif;
    PendingIntent notintent;

    /* RPC */
    LocBinder binder;
    public LocIF rpc_if;
    URL target_url;

    /* Our position data from last read */
    Location location;

    /* Battery info */
    BatteryState bat_rcvr;
    int bat_level;
    boolean charger;

    /* Prefs */
    Preferences prefs;

    @Override public void onCreate() {
		super.onCreate();
		System.out.println("GPS Service onCreate.");
	
	    setupHttpConnection();
	    
		setupLocationListener();
	
		setupAlarms();
	
		setupBattery();
	
		/* Get prefs */
		this.prefs = new Preferences(this);
		this.prefs.load();
		reconfigure();

		/* pripadne nastaveni u SSL konekci jen na CA ve vlastni databazi klicu */ 
		setupKeyStore(); // nutne volat az po nacteni "prefs" (coz vyse cinim)
		
		/* Create formatter */
		dateformatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SS'Z'");
		dateformatter.setTimeZone(new SimpleTimeZone(0, "UTC"));
	
		/* Create binder */
		this.binder = new LocBinder(this);
    }

    @Override public void onDestroy() {
		super.onDestroy();
		System.out.println("GPS Service onDestroy.");
		
		/* Remove alarms and unhook locator */
		this.am.cancel(this.amintent);
		this.am.cancel(this.tointent);
		this.lm.removeUpdates(this.ll);
		unregisterReceiver(this.recvr);
		unregisterReceiver(this.bat_rcvr);
		unregisterReceiver(this.recvTimeout);
	
		if (this.notman != null) {
		    /* Remove notification */
		    this.notman.cancelAll();
		}
    }


    @Override public IBinder onBind(Intent i) {
    	return this.binder;
    }

    @Override public boolean onUnbind(Intent i) {
		this.rpc_if = null;
		return false;
    }


    private void startLocator()	throws IllegalArgumentException {
		if (this.prefs.provider == 1)
		    this.lm.requestLocationUpdates(LocationManager.GPS_PROVIDER, 0, 0, this.ll);
		else
		    this.lm.requestLocationUpdates(LocationManager.NETWORK_PROVIDER, 0, 0, this.ll);
    }

    public void triggerUpdate() {
		if (this.prefs.continous_mode) {
		    /* If we are in continous mode just send last location */
		    locationUpdate();
		} 
		else {
		    /* Polling mode */
	
		    this.lm.removeUpdates(this.ll);
		    
		    try {
		    	startLocator();
		    }
		    catch (IllegalArgumentException e) {
		    	System.out.println("MyBrother: "+e.toString());
		    	return;
		    }
	
		    /* Start timeout alarm */
		    this.timeout = System.currentTimeMillis();
		    this.timeout += this.prefs.gps_timeout;
		    this.timeout += 100; /* delay a bit to avoid a race */
		    this.am.setRepeating(AlarmManager.RTC_WAKEUP, this.timeout, this.prefs.gps_timeout, this.tointent);
		    if (this.prefs.provider == 1)
		    	this.twiceTimeout = true;
		    else
		    	this.twiceTimeout = false;
		}
    }

    private void setupHttpConnection() {
    	this.httpclient = new DefaultHttpClient();
    	this.httpclient.setHttpRequestRetryHandler(new DefaultHttpRequestRetryHandler(1, false)); // retry jen 1x (napriklad je nedostupna sit) (pokud bych to neuvedl, tak defaultne je tusim retry 3x)
    	this.nameValuePairs = new ArrayList<NameValuePair>(2);
    }

    private void setupKeyStore() {
    	if(this.prefs.own_key_store) {
			try {
				KeyStore trustStore = KeyStore.getInstance(KeyStore.getDefaultType());
				FileInputStream instream = new FileInputStream(new File("/sdcard/security/cacerts.bks"));
                trustStore.load(instream, "changeit".toCharArray());
                instream.close();
                SSLSocketFactory socketFactory = new SSLSocketFactory(trustStore);
                Scheme sch = new Scheme("https", socketFactory, 443);
                this.httpclient.getConnectionManager().getSchemeRegistry().register(sch);
			}
			catch (KeyStoreException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch (NoSuchAlgorithmException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch (CertificateException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch (KeyManagementException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch (UnrecoverableKeyException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} 
    	}
    }

    private void setupLocationListener() {
		/* Set up the position listener */
		this.ll = new LocListen();
		this.lm = (LocationManager)getSystemService(Context.LOCATION_SERVICE);
    }

    private void setupNotif() {
		if (this.prefs.show_in_notif_bar) {
		    /* Show state in notif bar */
	
		    if (this.notman != null) {
			/* Already set up */
			return;
		    }
	
		    /* Set up a persistent notification */
		    this.notman = (NotificationManager)
			getSystemService(Context.NOTIFICATION_SERVICE);
		    this.notif = new Notification(R.drawable.notif_icon,
		    			  getString(R.string.app_name)+" "+getString(R.string.notifiWaitingForLocation),
						  System.currentTimeMillis());
		    this.notif.flags = Notification.FLAG_ONGOING_EVENT;
		    this.notintent = 
			PendingIntent.getActivity(this, 0, 
						  new Intent(this, MyBrother.class),0);
		    this.notif.setLatestEventInfo(this, getString(R.string.app_name),
		    		      getString(R.string.notifiWaitingForInitialLocation), 
						  notintent);
		    this.notman.notify(0, notif);
		}
		else {
		    if (this.notman != null) {
				/* Remove notification */
				this.notman.cancelAll();
				this.notman = null;
		    }	    
		}
    }

    private void setupAlarms() {
		/* Prepare alarm manager */
		this.am = (AlarmManager)getSystemService(Context.ALARM_SERVICE);
	
		/* Setup location update alarm */
		this.recvr = new LocAlarm();
		registerReceiver(this.recvr, 
				 new IntentFilter(LocAlarm.class.toString()),
				 null, null);
		Intent i = new Intent(LocAlarm.class.toString());
		this.amintent = PendingIntent.getBroadcast(this, 0, i, 0);
	
		/* Setup timeout alarm */
		this.recvTimeout = new LocTimeout();
		registerReceiver(this.recvTimeout, 
				 new IntentFilter(LocTimeout.class.toString()),
				 null, null);
		i = new Intent(LocTimeout.class.toString());
		this.tointent = PendingIntent.getBroadcast(this, 0, i, 0);
    }

    private void setupBattery() {
		/* Setup location update alarm */
		this.bat_rcvr = new BatteryState();
		registerReceiver(this.bat_rcvr, 
				 new IntentFilter(Intent.ACTION_BATTERY_CHANGED),
				 null, null);
    }
    

    public void doTimeout() {
		if (System.currentTimeMillis() > this.timeout) {
		    System.out.println("MyBrother: Doing timeout");
		    if (this.twiceTimeout) {
				System.out.println("MyBrother: Switching locator");
				this.twiceTimeout = false;
				this.timeout = 
				    System.currentTimeMillis() + this.prefs.gps_timeout;
				try {
				    this.lm.requestLocationUpdates(LocationManager.NETWORK_PROVIDER, 
								   0, 0, this.ll);
				}
				catch (IllegalArgumentException e) {
				    System.out.println("MyBrother(timeout): "
						       +e.toString());
				    return;
				}
		    }
		    else {
				/* Timeout reached */
				this.lm.removeUpdates(this.ll);
				this.am.cancel(this.tointent);
		    }
		}
    }

	
    public void reloadPrefs() {
    	this.prefs.load();
    	reconfigure();
    }

    private void reconfigure() {
		System.out.println("MyBrother doing reconfig");
		/* Update the request times */
		this.lm.removeUpdates(ll);
	
		/* Reset update alarms */
		this.am.setRepeating(AlarmManager.RTC_WAKEUP,
				     System.currentTimeMillis() + 1000,
				     this.prefs.update_interval, this.amintent);
	
		/* Fix notifs */
		setupNotif();
	
		/* Set URL */
		this.target_url = null;
		try {
		    this.target_url = new URL(this.prefs.target_url);
		}
		catch (MalformedURLException e) {
		    System.out.println("MyBrother: "+e.toString());
		    this.target_url = null;
		}
	
		/* For continous mode, start updating */
		if (this.prefs.continous_mode) {
		    try {
		    	startLocator();
		    }
		    catch (IllegalArgumentException e) {
				System.out.println("MyBrother: "
						   +"Can't start locator in continous mode: "
						   +e.toString());
				return;
		    }
		}
    }

    /* Send a request to the URL and post some data */
    protected void postLocation() {

		/* No url, don't do anything */
		if (this.target_url == null)
		    return;
		
		if(
				this.prefs.show_in_notif_bar &&
				this.prefs.http_resp_in_notif_bar
		) {
			this.setupNotif();
			if(
					this.notif != null &&
					waiting_for_first_send_data==1
			) {
			    this.notif.when = System.currentTimeMillis();
			    this.notif.setLatestEventInfo(
			    		this, 
						getString(R.string.app_name),
						getString(R.string.notifiWaitingForFirstSendDataToServer), 
						this.notintent
				);
			    this.notman.notify(0, this.notif);
			    waiting_for_first_send_data = 0;
			}
		}
		
		/* Build request data */

		++numberPoint;
		++numberPointSends;
		String numberPointText;
		if(numberPoint==1) {
			numberPointText="";
		}
		else {
			numberPointText=String.valueOf(numberPoint);
		}
	
		nameValuePairs.add(new BasicNameValuePair("latitude"+numberPointText, String.valueOf(this.location.getLatitude())));
		nameValuePairs.add(new BasicNameValuePair("longitude"+numberPointText, String.valueOf(this.location.getLongitude())));
		nameValuePairs.add(new BasicNameValuePair("accuracy"+numberPointText, String.valueOf(this.location.getAccuracy())));
	
		if (this.prefs.send_altitude) {
		    nameValuePairs.add(new BasicNameValuePair("altitude"+numberPointText, String.valueOf(this.location.getAltitude())));
		}
			
		if (this.prefs.send_provider) {
		    nameValuePairs.add(new BasicNameValuePair("provider"+numberPointText, this.location.getProvider()));
		}
	
		if (this.prefs.send_bearing) {
		    nameValuePairs.add(new BasicNameValuePair("bearing"+numberPointText, String.valueOf(this.location.getBearing())));
		}
	
		if (this.prefs.send_speed) {
		    nameValuePairs.add(new BasicNameValuePair("speed"+numberPointText, String.valueOf(this.location.getSpeed())));
		}
	
		if (this.prefs.send_time) {
		    Date date = new Date(this.location.getTime());
		    nameValuePairs.add(new BasicNameValuePair("time"+numberPointText, this.dateformatter.format(date)));
		}
	
		/* Add battery status if configured */
		if (this.prefs.send_batt_status) {
		    nameValuePairs.add(new BasicNameValuePair("battlevel"+numberPointText, String.valueOf(this.bat_level)));
		    if (this.charger) {
				nameValuePairs.add(new BasicNameValuePair("charging"+numberPointText, "1"));
		    }
		    else {
				nameValuePairs.add(new BasicNameValuePair("charging"+numberPointText, "0"));
		    }
		}
	
		/* Add secret if configured */
		if (this.prefs.secret != null) {
		    nameValuePairs.add(new BasicNameValuePair("secret"+numberPointText, this.prefs.secret));
		}
	
		/* Add device id */
		if (this.prefs.send_devid) {
		    TelephonyManager tm = (TelephonyManager)getSystemService(TELEPHONY_SERVICE);
		    nameValuePairs.add(new BasicNameValuePair("deviceid"+numberPointText, tm.getDeviceId()));
		}

		/* Add subscriber id */
		if (this.prefs.send_subscrid) {
		    TelephonyManager tm = (TelephonyManager)getSystemService(TELEPHONY_SERVICE);
		    nameValuePairs.add(new BasicNameValuePair("subscriberid"+numberPointText, tm.getSubscriberId()));
		    if(tm.getPhoneType()==TelephonyManager.PHONE_TYPE_GSM) {
		    	nameValuePairs.add(new BasicNameValuePair("networkoperatorname"+numberPointText, tm.getNetworkOperatorName()));
		    	nameValuePairs.add(new BasicNameValuePair("networktype"+numberPointText, String.valueOf(tm.getNetworkType())));
		    }
		}
		
		if(numberPointSends>=this.prefs.send_points_together || MyBrother.send_to_server_now==1) {
			
			numberPointSends = 0;

			String serverinfo="";
			
			HttpPost httppost = new HttpPost(this.target_url.toString());
	
		    /* HTTP Basic Authorization */
			if (this.prefs.http_user != null && this.prefs.http_password != null) {
				// 1. varianta - je asi poctiva - prvni pozadavek je bez jmena a hesla a az si server vyzada autorizaci bude nasledovat pozadavek s uzivatelem a heslem (definice nize by vsak mela byt pripadne ve funkci setupHttpConnection())
		        //httpclient.getCredentialsProvider().setCredentials(
		        //        new AuthScope(AuthScope.ANY_HOST, AuthScope.ANY_PORT),
		        //        new UsernamePasswordCredentials(this.prefs.http_user, this.prefs.http_password)
		        //);
				// 2. varianta - poctiva - kdy jen u prvniho pozadavku je pozadavek bez jmena a hesla a az si server vyzada autorizaci bude nasledovat pozadavek s uzivatelem a heslem a u dalsich pozadavku bude jiz vzdy uzivatel a heslo
				//    - tohle jsem rozracoval u bigbrothergps-snapshot-mara03-httpclient-preemptive_authentication-for_android_2.2-TEST
				//      dle stranky zde http://hc.apache.org/httpcomponents-client-ga/tutorial/html/authentication.html,
				//      ale jsou tam jine zaludnosti a zatim to delam jinak (viz nize)
				// 3. varianta - mozna to ma nejakou zaludnost - natvrdo do hlavicky vlozim informace pro basic autorizaci a pozadavek je tak jen jeden
				String http_userpassword = this.prefs.http_user + ":" + this.prefs.http_password; 
				httppost.setHeader("Authorization", "basic " + Base64.encodeToString(http_userpassword.getBytes(),Base64.NO_WRAP));
			}
	
			/* Connect and write */
			HttpResponse response;
			try {
				httppost.setEntity(new UrlEncodedFormEntity(nameValuePairs));
				response = this.httpclient.execute(httppost);
	
				HttpEntity entity = response.getEntity();
				if (entity != null) {
					serverinfo=EntityUtils.toString(entity); // tohle muze asi pripadne nahradit radek nize (pokud budes chti s obsahem co je vracen pracovat)
					if(serverinfo.length()>100) {
						serverinfo=serverinfo.substring(0,100);
					}
					//entity.consumeContent(); // zajisti asi konzumaci obsahu s ukoncenim pozadavku - diky tomu funguje keep-alive (kdyz to tu nebylo, keep-alive nefungoval - nestacilo jen posunout definic httpclient nahoru)
			    }
			}
			catch (ClientProtocolException e) {
				// TODO Auto-generated catch block
			}
			catch (IOException e) {
			    System.out.println("MyBrother: "+e.toString());
			    if (this.rpc_if != null)
				this.rpc_if.onError(e.toString());
			    return;
			}
	
			numberPoint = 0;
			nameValuePairs = null;
			nameValuePairs = new ArrayList<NameValuePair>(2);
			
			System.out.println("MyBrother sent HTTP poke");
			
			/* Set notification if we have it */
			if (
					this.notif != null &&
					this.prefs.show_in_notif_bar &&
					this.prefs.http_resp_in_notif_bar
			) {
			    this.notif.when = System.currentTimeMillis();
			    this.notif.setLatestEventInfo(
			    		this, 
						getString(R.string.app_name),
						"SERVER: "+serverinfo, 
						this.notintent
				);
			    this.notman.notify(0, this.notif);
			}
		  			
		}

    }

    private void locationUpdate() {
		/* Location update triggered */
		Location loc = this.location;
		if (loc == null)
		    return;
	
		/* Change notification */
		if (this.prefs.show_in_notif_bar && !this.prefs.http_resp_in_notif_bar) {
		    this.setupNotif();
		    String txt = loc.getLatitude()+", "
			+loc.getLongitude()+", "
			+(int)loc.getAccuracy()+"m";
		    this.notif.when = System.currentTimeMillis();
		    this.notif.setLatestEventInfo(GPS.this, 
						      getString(R.string.app_name),
						      txt, GPS.this.notintent);
		    this.notman.notify(0, GPS.this.notif);
		}
		
		/* Call to UI */
		if (this.rpc_if != null) {
		    this.rpc_if.onLocation(loc.getProvider(), loc,
					   this.bat_level,
					   this.charger);
		}
		
		/* Post to server */
		this.postLocation();
    }

    /**************************************************************************
     * Helper classes for timed locator interaction
     *************************************************************************/
    class LocAlarm extends BroadcastReceiver {
		@Override public void onReceive(Context ctx, Intent i) {
		    System.out.println("MyBrother: Alarm!");
		    GPS.this.triggerUpdate();
		}
    }

    class LocTimeout extends BroadcastReceiver {
		@Override public void onReceive(Context ctx, Intent i) {
		    if (GPS.this.prefs.continous_mode) {
		    	System.out.println("MyBrother: Ignored timeout");
		    }
		    else {
		    	System.out.println("MyBrother: Received Timeout!");
		    	GPS.this.doTimeout();
		    }
		}
    }

    class BatteryState extends BroadcastReceiver {
		@Override public void onReceive(Context ctx, Intent i) {
		    float level;
		    level = i.getIntExtra("level", 0);
		    level /= i.getIntExtra("scale", 100);
		    GPS.this.bat_level = (int)(level*100);
		    GPS.this.charger = i.getIntExtra("plugged",1)!=0;
		    System.out.printf("MyBrother: Battery state change: %d%% %b\n",
				      GPS.this.bat_level, GPS.this.charger);
		}
    }

    class LocListen implements LocationListener {
		@Override public void onProviderDisabled(String prov) {
		    System.out.println("MyBrother ProviderDisabled: "+prov);
		    if (GPS.this.prefs.continous_mode) 
		    	GPS.this.startLocator();
		    else {
		    	/* Stop waiting for timeout since this locator is off */
		    	GPS.this.timeout = System.currentTimeMillis()-1;
		    	GPS.this.doTimeout();
		    }
		}
	
		@Override public void onProviderEnabled(String prov) {
		    System.out.println("MyBrother ProviderEnabled: "+prov);
		    if (GPS.this.prefs.continous_mode) 
		    	GPS.this.startLocator();
		    else
		    	GPS.this.doTimeout();	    
		}

		@Override public void onStatusChanged(String prov, int stat, Bundle xtra) {
		    System.out.println("MyBrother Status change: "+prov+" -> "+stat);
		    if (GPS.this.rpc_if != null)
		    	GPS.this.rpc_if.onStateChange(prov, stat);
	
		    if (GPS.this.prefs.continous_mode) 
		    	GPS.this.startLocator();
		    else
		    	GPS.this.doTimeout();
		}
	
		@Override public void onLocationChanged(Location loc) {
		    System.out.println("MyBrother got loc from "+loc.getProvider());
		    GPS.this.location = loc;
	
		    if (!GPS.this.prefs.continous_mode) {
				/* Stop waiting for locations. Will be restarted by alarm */
				GPS.this.lm.removeUpdates(GPS.this.ll);
				GPS.this.am.cancel(GPS.this.tointent);
		
				GPS.this.locationUpdate();
		    }
		}
    }
}
