package net.chlup.mybrother;

import java.util.*;

import java.text.SimpleDateFormat;

import android.app.Activity;
import android.os.Bundle;
import android.os.IBinder;
import android.content.Intent;
import android.content.ComponentName;
import android.content.ServiceConnection;

import android.view.Menu;
import android.view.MenuItem;

import android.view.View;
import android.widget.TextView;
import android.widget.Button;

import android.location.*;

import net.chlup.android.Helper;

public class MyBrother extends Activity {
	Bundle state;

	/* Prefs */
	Preferences prefs;
	
	public static int send_to_server_now = 0;

	/* References to our service */
	Intent srvint;
	ComponentName srv;
	Con servicecon;
	LocBinder binder;

	/* Our UI components */
	private TextView time;
	private TextView prov, lat, lon, alt, acc;
	private TextView brg, spd;
	private TextView batlev, chrgr;
	private TextView log;

	SimpleDateFormat dateformatter;

	/** Called when the activity is first created. */
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		System.out.println("onCreate called");
		this.state = savedInstanceState;

		/* Get prefs */
		this.prefs = new Preferences(this);
		this.prefs.load();

		startTheGPS();

		startUI();
	}

	@Override
	public void onDestroy() {
		super.onDestroy();
		System.out.println("onDestroy called");

		/* Disconnect from service */
		unbindService(this.servicecon);
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		menu.add(0, 0, 0, R.string.mainUpdateLocation);
		menu.add(0, 1, 1, R.string.mainChangelog);
		menu.add(0, 100, 100, R.string.mainSettings).setIcon(
				android.R.drawable.ic_menu_preferences);
		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case 0:
			if (this.binder != null)
				this.binder.triggerUpdate();
			break;
		case 1:
			Intent changelog = new Intent(this, Changelog.class);
			startActivity(changelog);
			break;
		case 100:
			Intent prefA = new Intent(this, PreferencesActivity.class);
			startActivityForResult(prefA, 0);
			break;
		}

		return true;
	}

	@Override
	public void onActivityResult(int req, int res, Intent i) {
		/* Get prefs */
		this.prefs.load();

		/* Notify the service that preferences might have changed */
		if (this.binder != null)
			this.binder.updatePrefs();
	}

	private void startTheGPS() {
		/* Start GPS service */
		this.srvint = new Intent(this, GPS.class);
		this.srv = startService(this.srvint);
		this.servicecon = new Con();
		boolean bind = bindService(this.srvint, this.servicecon, 0);
		if (this.srv == null && bind) {
			Helper.ok_dialog(this, "Service", "Failed starting GPS service");
			finish();
		}
	}

	private void startUI() {
		/* Create the UI */
		setContentView(R.layout.main);

		/* Date/time formatting */
		this.dateformatter = new SimpleDateFormat("HH:mm:ss.SS");

		/* Lookup our text fields */
		this.time = (TextView) findViewById(R.id.main_time);

		this.prov = (TextView) findViewById(R.id.main_provider);
		this.lat = (TextView) findViewById(R.id.main_latitude);
		this.lon = (TextView) findViewById(R.id.main_longitude);
		this.alt = (TextView) findViewById(R.id.main_altitude);
		this.acc = (TextView) findViewById(R.id.main_accuracy);

		this.brg = (TextView) findViewById(R.id.main_bearing);
		this.spd = (TextView) findViewById(R.id.main_speed);

		this.batlev = (TextView) findViewById(R.id.main_bat_level);
		this.chrgr = (TextView) findViewById(R.id.main_bat_charger);

		this.log = (TextView) findViewById(R.id.main_log);

		/* Hook the button */
		Button btn;

		/* end button */
		btn = (Button) findViewById(R.id.main_stop);
		if (btn != null)
			btn.setOnClickListener(new View.OnClickListener() {
				public void onClick(View v) {
					stopService(MyBrother.this.srvint);
					MyBrother.this.finish();
				}
			});

		/* update button */
		btn = (Button) findViewById(R.id.main_triggerupdate);
		if (btn != null)
			btn.setOnClickListener(new View.OnClickListener() {
				public void onClick(View v) {
					if (MyBrother.this.binder != null) {
						send_to_server_now = 1;
						MyBrother.this.binder.triggerUpdate();
						send_to_server_now = 0;
					}
				}
			});
	}

	void logError(String err) {
		/* Add error message at top. */
		this.log.setText(err + "\n" + this.log.getText());

		/* Keep only last 10 */
		String oldlog = this.log.getText().toString();
		String log[] = oldlog.split("\n");
		if (log.length > 10) {
			StringBuffer newlog = new StringBuffer();
			for (int i = 0; i < 10; i++)
				newlog.append(log[i] + "\n");
			this.log.setText(newlog);
		}
	}

	/************************************************************************
	 * Helper classes for communication with the service
	 ************************************************************************/

	/* Wrapper so we get connected to the service on bind */
	class Con implements ServiceConnection {
		@Override
		public void onServiceConnected(ComponentName name, IBinder service) {
			LocBinder lb = (LocBinder) service;
			CallBackIF cb = new CallBackIF();

			/* Read last position from locator */
			cb.onLocation("init", lb.getLocation(), lb.getBattery(), lb
					.getCharger());

			/* Bind for updates */
			MyBrother.this.binder = lb;
			lb.setCallback(new CallBackIF());
		}

		@Override
		public void onServiceDisconnected(ComponentName name) {
		};
	}

	/* The actual connection interface class */
	class CallBackIF implements LocIF {
		@Override
		public void onError(String err) {
			MyBrother.this.logError(err);
		}

		@Override
		public void onStateChange(String prov, int state) {
		}

		@Override
		public void onLocation(String prov, Location loc, int bat_level,
				boolean charger) {
			if (loc == null)
				return;

			Double latitude = new Double(loc.getLatitude());
			Double longitude = new Double(loc.getLongitude());
			Integer accuracy = new Integer((int) loc.getAccuracy());
			Double altitude = new Double(loc.getAltitude());
			Double bearing = new Double(loc.getBearing());
			Double speed = new Double(loc.getSpeed());

			Date date = new Date(loc.getTime());
			String df = MyBrother.this.dateformatter.format(date);

			/* Set format */
			int cf;
			switch (MyBrother.this.prefs.coordinate_format) {
			case 2:
				cf = Location.FORMAT_MINUTES;
				break;
			case 3:
				cf = Location.FORMAT_SECONDS;
				break;
			default:
				cf = Location.FORMAT_DEGREES;
				break;
			}

			MyBrother.this.time.setText(df);
			MyBrother.this.prov.setText(loc.getProvider());
			MyBrother.this.lat.setText(Location.convert(latitude, cf));
			MyBrother.this.lon.setText(Location.convert(longitude, cf));
			MyBrother.this.alt.setText(altitude.toString());
			MyBrother.this.acc.setText(accuracy.toString());
			MyBrother.this.brg.setText(bearing.toString());
			MyBrother.this.spd.setText(speed.toString());
			MyBrother.this.batlev.setText((new Integer(bat_level))
					.toString());
			if (charger)
				MyBrother.this.chrgr.setText("(charging)");
			else
				MyBrother.this.chrgr.setText("");
		}
	}
}
